/* Copyright (c) 2005 Dirk Jagdmann <doj@cubic.org>

This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you
       must not claim that you wrote the original software. If you use
       this software in a product, an acknowledgment in the product
       documentation would be appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and
       must not be misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
       distribution. */

// $Header: /home/doj/code/doj/sndfile.hpp,v 1.1 2005-04-15 00:08:57 doj Exp $

#ifndef SNDFILE__HPP
#define SNDFILE__HPP

#include <sndfile.h>
#include <string>
#include <iostream>

namespace doj
{

  namespace {
    inline sf_count_t read_helper(SNDFILE *sf, short *p, sf_count_t f)
    {
      return sf_readf_short(sf, p, f);
    }

    inline sf_count_t read_helper(SNDFILE *sf, int *p, sf_count_t f)
    {
      return sf_readf_int(sf, p, f);
    }

    inline sf_count_t read_helper(SNDFILE *sf, float *p, sf_count_t f)
    {
      return sf_readf_float(sf, p, f);
    }

    inline sf_count_t read_helper(SNDFILE *sf, double *p, sf_count_t f)
    {
      return sf_readf_double(sf, p, f);
    }
  }

  /**
     load an audio file with libsndfile. http://www.mega-nerd.com/libsndfile

     The complete file is read into memory which is allocated with
     new[]. The caller has to free this memory.

     @param filename file to load into memory
     @param samplerate out param which describes the samplerate of filename
     @param channels out param which describes the numer of channels in filename
     @param size the size of the allocated memory block in frames/samples. This number is channels*length of sample.
     @param dummy not used parameter needed for template magic. This parameter has to be the same type as the loaded sample will have.

     @return a pointer to the samples upon success, 0 upon error
  */
  template <typename T>
  T* sndfile_read(const std::string& filename, int& samplerate, int& channels, unsigned& size, T *dummy)
  {
    SF_INFO info;
    SNDFILE *sf=sf_open(filename.c_str(), SFM_READ, &info);
    if(!sf)
      {
	std::cerr << "could not open " << filename << " with libsndfile\n";
	return 0;
      }

    samplerate=info.samplerate;
    channels=info.channels;
    size=info.frames*channels;
    T *buf=0;
    try {
      buf=new T[size];
      if(read_helper(sf, buf, info.frames) != info.frames)
	std::clog << "could not read all frames from " << filename << std::endl;
    }
    catch (std::bad_alloc&)
      {
	std::cerr << filename << " is too large to be read into (free) memory\n";
      }

    if(sf_close(sf) != 0)
      std::clog << sf_strerror(sf) << std::endl;
    return buf;
  }

}

#endif
